# encoding: utf-8
# 1^k + 2^k + …… + n^k をベルヌーイ数利用により算出

# 組み合わせパターン数(二項係数)を算出
def nck(n, k)
    kk = (1..k).to_a  # 1, 2, 3, …… k を配列にする
    c = Rational(1)
    for i in (n-k+1)..n
        c *= i
        c /= kk.pop
    end
    return c.to_i
end

# n番目までの一連のベルヌーイ数を算出
def bern1(num)
    bn = []  # この配列にベルヌーイ数を入れる
    bn[0] = 1
    for n in 1..num
        s = Rational(0)
        for i in 0..(n-1)
            s = s + nck(n+1, i) * bn[i]
        end
        bn[n] = -Rational(1, n+1) * s
    end
    return bn
end

# 1^k + 2^k + …… + n^k を算出：ベルヌーイ数を利用
def bsum(n, k)
    bn = bern1(k)
    n += 1
    s = Rational(0)
    for i in 0..k
        s = s + nck(k+1, i) * bn[i] * n**(k-i+1)
    end
    s = s / (k+1)
    return s.to_i
end

# 1^k + 2^k + …… + n^k を算出：通常の方法
def nsum(n, k)
    s = 0
    for i in 1..n
        s = s + i**k
    end
    return s
end

# テスト表示
for k in 2..4
    for n in 3..8
        printf("sum of '1^%d--%d^%d'\t%d\t%d\n", k, n, k,
            bsum(n,k), nsum(n,k))
    end
    printf("\n")
end
