# binomer.rb for binomial expansion  (encoding: utf-8)  2015/12/26
  # Copyright (C) T. Yoshiizumi, 2015 All rights reserved.
require "bigdecimal"

module Bnm
  def self.fact(n)
    f = 1
    for i in 1..n
      f *= i
    end
    return f
  end

  def self.ncr(n, r)
    nn = 1
    dd = 1
    for i in 1..r
      nn *= (n-i+1)
      dd *= i
    end
    ncr = Rational(nn, dd)
    ncr = ncr.numerator  if ncr.denominator == 1
    return [nn, dd, ncr]
  end

  def self.dcmp(arg1, arg2)
    arg1 = arg1.to_f  if arg1.class == Rational
    arg2 = arg2.to_f  if arg2.class == Rational
    acl = [arg1.class, arg2.class]
    s1 = (acl[0] == BigDecimal) ? arg1.to_s("F") : arg1.to_s
    s2 = (acl[1] == BigDecimal) ? arg2.to_s("F") : arg2.to_s
    b1 = s1.bytes
    b2 = s2.bytes
    ary = []
    for i in 0...b1.size
      break  unless b2[i]
      if b1[i] != b2[i]
        ary = [i, b1[i].chr, b2[i].chr]
        break
      end
    end
    if ary.size > 0
      str = sprintf("different position: %d (in %d & %d) | %c:%c\n%s\n\n%s\n",
          ary[0], s1.size, s2.size, ary[1], ary[2], s1, s2)
    else
      str = sprintf("can't find the difference (in %d & %d)\n%s\n\n%s\n",
          s1.size, s2.size, s1, s2)
    end
    return str
  end
end

class Binomer
  def r2i(val)  # Rational to Integer
    if val.class == Rational and val.denominator == 1
      val = val.numerator
    end
    return val
  end
  private :r2i

  def inner_substr(str)
    chrs = %w(c a x)
    for i in 0...chrs.size
      chr = chrs[i]
      str = str.sub(/^#{chr}([^0-9A-Za-z_])/, "tary[i][#{i}]\\1")
      str = str.sub(/([^0-9A-Za-z_$@])#{chr}$/, "\\1tary[i][#{i}]")
      str = str.gsub(/([^0-9A-Za-z_$@])#{chr}([^0-9A-Za-z_])/,
          "\\1tary[i][#{i}]\\2")
    end
    return str
  end
  private :inner_substr

  def initialize(n=nil, r=nil, xst=nil)
    @bary = []
    if n.class == Array
      self.from_a(n)
    elsif n != nil
      expansion(n, r, xst)
    end
  end

  def expansion(n, r=nil, xst=nil)
    unless n
      return self
    end
    r_org = r
    xst_org = xst
    if n.class != Rational
      n = Rational(n)
    end
    if r == nil
      if n.denominator == 1 and n >= 0
        r = n
      else
        r = Rational(10)
      end
    else
      r = Rational(r)
    end
    if r.denominator != 1
      STDERR.puts "expansion(): the 2'nd argument '#{r_org}' " +
          "should be an integer value."
      return self
    end
    xst = Rational(1)  unless xst
    xst = Rational(xst)  if xst.class != Rational
    if xst.denominator != 1
      STDERR.puts "expansion(): the 3'nd argument '#{xst_org}' " +
          "should be an integer value."
      return self
    end
    minus = false
    if xst < 0
      minus = true
      xst = -1 * xst
    end
    a = n
    x = Rational(0)
    ary = [[Rational(1), a, x]]
    ncr = Rational(1)
    for i in 1..r
      break  if i == (n + 1)
      a -= 1
      x += 1
      ncr *= (n - i + 1)
      ncr /= i
      ncr *= -1  if minus
      ary << [ncr, a, x * xst]
    end
    ary.pop  if !(n.denominator == 1 and n >= 0) and ary.size > 1
    @bary = ary
    return self
  end
  alias :expansion! :expansion
  alias :exp :expansion
  alias :exp! :expansion

  def coefficient()
    ary = []
    @bary.each do |c, a, x|
      ary << c
    end
    return ary
  end
  alias :coeff :coefficient

  def to_a()
    ary = []
    @bary.each do |aa|
      ary << aa.dup
    end
    return ary
  end

  def from_a(ary)
    ary2 = []
    ary.each do |aa|
      aa2 = []
      aa.each {|e|  aa2<<Rational(e)}
      if !(aa2[2].class == Rational and aa2[2].denominator == 1)
        STDERR.puts "from_a(): each array's 3'nd element " +
            "\"#{aa[2].inspect}\" should be an integer."
        break
      end
      if ary2.size > 0 and aa2[2] <= ary2.last[2]
        STDERR.puts "from_a(): 3'nd element " +
            "\"#{aa[2].inspect}\" should be greater than the previous one."
        break
      end
      ary2 << aa2  if aa2[0] != 0
    end
    if ary2.size == ary.size
      @bary = ary2
    end
    return self
  end
  alias :from_a! :from_a

  def cax(type = nil)
    ary = []
    @bary.each do |aa|
      ary << [r2i(aa[0]), r2i(aa[1]), r2i(aa[2])]
    end
    sary = []
    ary.each do |ncr, a, x|
      if ncr == 1
        str = ""
      elsif ncr == -1
        str = "-"
      else
        str = sprintf("%s", ncr)
      end
      if a == 0
        if str =~ /[^0-9)]$|^$/
          str += "1"
        end
      elsif a == 1
        str += "a"
      else
        if [Fixnum, Bignum].include?(a.class)
          str += sprintf("a^%s", a)
        else
          str += sprintf("a^(%s)", a)
        end
      end
      if x == 0
        if str =~ /[^0-9)]$|^$/
          str += "1"
        end
      elsif x == 1
        str = str.sub(/1$/, "")  if str =~ /^1$|[^\d]1$/
        str += "x"
      else
        str = str.sub(/1$/, "")  if str =~ /^1$|[^\d]1$/
        str += sprintf("x^%s", x)
      end
      sary << str
    end
    type = type.to_s
    if type =~ /^string(\d*)$/i
      sn = $1
      n = (sn != '') ? (sn.to_i - 1) : -1
      return sary[0..n].join(" + ").gsub(/ \+ \-/, " - ")
    end
    return sary
  end

  def cx(type = nil)
    ary = []
    @bary.each do |aa|
      ary << [r2i(aa[0]), r2i(aa[1]), r2i(aa[2])]
    end
    sary = []
    ary.each do |ncr, a, x|
      if ncr == 1
        str = ""
      elsif ncr == -1
        str = "-"
      else
        str = sprintf("%s", ncr)
      end
      if x == 0
        if str =~ /[^0-9)]$|^$/
          str += "1"
        end
      elsif x == 1
        str = str.sub(/1$/, "")  if str =~ /^1$|[^\d]1$/
        str += "x"
      else
        str = str.sub(/1$/, "")  if str =~ /^1$|[^\d]1$/
        str += sprintf("x^%s", x)
      end
      sary << str
    end
    type = type.to_s
    if type =~ /^string(\d*)$/i
      sn = $1
      n = (sn != '') ? (sn.to_i - 1) : -1
      return sary[0..n].join(" + ").gsub(/ \+ \-/, " - ")
    end
    return sary
  end

  def integral()
    tary = self.to_a
    if tary[0][2] < 0
      STDERR.puts "integral(): the series \"x^n\"'s n should be " +
          "equal to 0 or greater!"
      return
    end
    for i in 0...tary.size
      xst = tary[i][2] + 1
      tary[i][2] = xst
      ncr = tary[i][0]
      if xst != 1
        tary[i][0] = Rational(ncr) / xst
      end
    end
    return Binomer.new(tary)
  end
  alias :inte :integral

  def integral!()
    bnr = self.integral()
    @bary = bnr.to_a  if bnr.class == Binomer
    return self
  end
  alias :inte! :integral!

  def differential()
    tary = self.to_a
    if tary[0][2] < 0
      STDERR.puts "differential(): the series \"x^n\"'s n should be " +
          "equal to 0 or greater!"
      return
    end
    for i in 0...tary.size
      xst = tary[i][2]
      if xst < 1
        tary[i] = nil
        next
      end
      tary[i][2] = xst - 1
      tary[i][0] = tary[i][0] * xst
    end
    tary.delete(nil)
    return Binomer.new(tary)
  end
  alias :diff :differential

  def differential!()
    bnr = self.differential()
    @bary = bnr.to_a  if bnr.class == Binomer
    return self
  end
  alias :diff! :differential!

  def calc(arg1, arg2=nil)
    acl = [arg1.class, arg2.class]
    if acl.include?(BigDecimal)
      scale = BigDecimal.new("0").precs[1]
      if arg2 == nil
        scale = arg1.precs[1]
        va = 1
        vx = arg1
      else
        if acl[0] == BigDecimal
          scale = arg1.precs[1]
        end
        if acl[1] == BigDecimal
          scale = arg2.precs[1]  if arg2.precs[1] > scale
        end
        va = (acl[0] != BigDecimal) ? BigDecimal.new(arg1.to_s) : arg1
        vx = (acl[1] != BigDecimal) ? BigDecimal.new(arg2.to_s) : arg2
      end
      sum = BigDecimal.new("0")
      @bary.each do |ncr, a, x|
        nn = BigDecimal.new(ncr.numerator)
        dd = BigDecimal.new(ncr.denominator)
        val = nn.mult(vx.power(x, scale), scale)
        val = val.mult(va.power(a, scale), scale)  if va != 1
        val = val.div(dd, scale)
        sum = sum.add(val, scale)
      end
      return sum
    else  # not BigDecimal => Float
      if arg2 == nil
        va = 1.0
        vx = arg1.to_f
      else
        va = arg1.to_f
        vx = arg2.to_f
      end
      sum = 0.0
      @bary.each do |ncr, a, x|
        nn = ncr.numerator.to_f
        dd = ncr.denominator.to_f
        val = nn * vx**x
        val *= va**a  if va != 1.0
        val = val / dd
        sum += val
      end
      return sum
    end
  end

  def inner(str)
    str2 = inner_substr(str)
    if str == str2
      STDERR.puts "inner(): '#{str}' is undue"
      return nil
    end
    tary = self.to_a
    for i in 0...tary.size
      eval(str2)
    end
    return Binomer.new(tary)
  end

  def inner!(str)
    bnr = self.inner(str)
    @bary = bnr.to_a  if bnr.class == Binomer
    return self
  end

  def dup()
    w = Binomer.new
    w.from_a(@bary)
    return w
  end
  alias :copy :dup

  def +(other)
    if [Fixnum, Bignum, Float, Rational].include?(other.class)
      tary = self.to_a
      if tary[0][2] == 0
        tary[0][0] += other
      else
        tary.unshift([other, 0, 0])
      end
      tary.shift  if tary[0][0] == 0
      return Binomer.new(tary)
    end
    bary1 = self.to_a
    bary2 = other.to_a
    ary = []
    while bary1.size > 0 or bary2.size > 0
      if bary1.size == 0
        ary << bary2.shift
        next
      end
      if bary2.size == 0
        ary << bary1.shift
        next
      end
      if bary1[0][2] == bary2[0][2]
        bary1[0][0] = bary1[0][0] + bary2[0][0]
        wary = bary1.shift
        ary << wary  if wary[0] != 0
        bary2.shift
      elsif bary1[0][2] < bary2[0][2]
        ary << bary1.shift
      else
        ary << bary2.shift
      end
    end
    return Binomer.new(ary)
  end

  def -(other)
    if [Fixnum, Bignum, Float, Rational].include?(other.class)
      tary = self.to_a
      if tary[0][2] == 0
        tary[0][0] -= other
      else
        tary.unshift([-1 * other, 0, 0])
      end
      tary.shift  if tary[0][0] == 0
      return Binomer.new(tary)
    end
    bary1 = self.to_a
    bary2 = other.to_a
    ary = []
    while bary1.size > 0 or bary2.size > 0
      if bary1.size == 0
        bary2[0][0] *= -1
        ary << bary2.shift
        next
      end
      if bary2.size == 0
        ary << bary1.shift
        next
      end
      if bary1[0][2] == bary2[0][2]
        bary1[0][0] = bary1[0][0] - bary2[0][0]
        wary = bary1.shift
        ary << wary  if wary[0] != 0
        bary2.shift
      elsif bary1[0][2] < bary2[0][2]
        ary << bary1.shift
      else
        bary2[0][0] *= -1
        ary << bary2.shift
      end
    end
    return Binomer.new(ary)
  end

  def *(other)
    if [Fixnum, Bignum, Float, Rational].include?(other.class)
      tary = self.to_a
      for i in 0...tary.size
        tary[i][0] *= other
      end
      return Binomer.new(tary)
    end
    if other.class == String
      n = nil
      if other =~ /^x$/i
        n = 1
      elsif other =~ /^x\^(\d+)$/i
        n = $1.to_i
      end
      unless n
        STDERR.puts "'*': the argument '#{other}' is undue."
        return
      end
      return self.inner("x += #{n}")
    end
    bary1 = self.to_a
    bary2 = other.to_a
    if bary1[0][2] < 0 or bary2[0][2] < 0
      STDERR.puts "'*': the target series \"x^n\"'s n should be " +
          "equal to 0 or greater!"
      return
    end
    ary = []
    for i in 0...bary1.size
      for j in 0...bary2.size
        x1 = bary1[i][2]
        x2 = bary2[j][2]
        x = x1 + x2
        ary[x] = []  unless ary[x]
        ary[x] << bary1[i][0] * bary2[j][0]
      end
    end
    ary2 = []
    for x in 0...ary.size
      next  unless ary[x]
      sum = Rational(0)
      ary[x].each do |aa|
        sum = sum + aa
 sum
      end
      ary2 << [sum, Rational(1), Rational(x)]  if sum != Rational(0)
    end
    return Binomer.new(ary2)
  end

  def power(n)
    n = r2i(n)
    if n.class != Fixnum
      STDERR.puts "power(): the argument should be a natural number"
      return self
    end
    if n == 0
      return Binomer.new([[1, 1, 0]])
    end
    if n == 1
      return self.copy
    end
    if @bary[0][2] < 0
      STDERR.puts "power(): the series \"x^n\"'s n should be " +
          "equal to 0 or greater!"
      return
    end
    bnr = self.copy
    base = self.copy
    for i in 2..n
      base = base.copy * bnr
    end
    return base
  end

  def reciprocalx(n = 10, name = "reciprocalx")
    ary = self.to_a
    if ary[0][2] < 0
      STDERR.puts "#{name}(): the series \"x^n\"'s n should be " +
          "equal to 0 or greater!"
      return
    end
    if ary[0][2] == 0
      ary[0][0] -= 1
      ary.shift  if ary[0][0] == 0
    else
      ary.unshift([-Rational(1), Rational(1), Rational(0)])
    end
    for i in 0...ary.size
      ary[i][0] *= -1
    end
    if ary[0][2] == 0 and ary[0][0] != -1
      STDERR.puts "#{name}(): can't be divided by this object!"
      return
    end
    elm = Binomer.new(ary)
    res = Binomer.new([[1, 0, 0]])
    res = res + elm
    for i in 2..n
      res = res + elm.copy.power(i)
    end
    return res
  end
  alias :recix :reciprocalx

  def reciprocal(n = 10)
    bnr = self.reciprocalx(n, "reciprocal")
    return bnr.round(self)
  end
  alias :reci :reciprocal

  def round(arg)
    if arg.class == Fixnum or arg.class == Bignum
      x = arg
    elsif arg.class == Binomer
      x = arg.to_a.last[2]
    else
      STDERR.puts "round(): the argument '#{arg.inspect}' is undue"
      return nil
    end
    tary = self.to_a
    for i in 0...tary.size
      tary[i] = nil  if tary[i][2] > x
    end
    tary.delete(nil)
    return Binomer.new(tary)
  end
  alias :cut :round

  def /(other)
    if [Fixnum, Bignum, Float, Rational].include?(other.class)
      tary = self.to_a
      for i in 0...tary.size
        tary[i][0] /= other
      end
      return Binomer.new(tary)
    end
    if other.class == String
      n = nil
      if other =~ /^x$/i
        n = 1
      elsif other =~ /^x\^(\d+)$/i
        n = $1.to_i
      end
      unless n
        STDERR.puts "'/': the argument '#{other}' is undue."
        return
      end
      return self.inner("x -= #{n}")
    end
    bnr1 = self.copy
    bnr2 = other.reciprocal
    res = (bnr1 * bnr2).round(bnr1)
    return res
  end

  def unshift(arg)
    tary = self.to_a
    if tary[0][2] < 0
      STDERR.puts "unshift(): the first \"x^n\"'s n is less than 0!"
      return
    end
    if tary[0][2] == 0
      STDERR.puts "unshift(): This series already has the constant term!"
      return
    end
    tary.unshift([arg, 0, 0])
    return Binomer.new(tary)
  end

  def first()
    res = nil
    res = @bary.first.dup  if @bary.class == Array and @bary.size > 0
    return res
  end

  def last()
    res = nil
    res = @bary.last.dup  if @bary.class == Array and @bary.size > 0
    return res
  end


  def self.e(n = 10)
    ary = []
    for x in 0...n
      ary << [Rational(1, Bnm::fact(x)), 0, x]
    end
    return Binomer.new(ary)
  end

  def self.log(n = 10)
    pm = 1
    ary = []
    for x in 1..n
      ary << [pm * Rational(1, x), 0, x]
      pm *= -1
    end
    return Binomer.new(ary)
  end

  def self.sin(n = 10)
    pm = 1
    ary = []
    for i in 1..n
      x = i * 2 - 1
      ary << [pm * Rational(1, Bnm::fact(x)), 0, x]
      pm *= -1
    end
    return Binomer.new(ary)
  end

  def self.cos(n = 10)
    pm = 1
    ary = []
    for i in 0...n
      x = i * 2
      ary << [pm * Rational(1, Bnm::fact(x)), 0, x]
      pm *= -1
    end
    return Binomer.new(ary)
  end
end
